@@ -1,4 +1,4 @@
-// Copyright (c) 2015-2016 Josh Blum
+// Copyright (c) 2015-2017 Josh Blum
 // Copyright (c) 2016-2016 Bastille Networks
 // SPDX-License-Identifier: BSL-1.0
 
@@ -10,19 +10,16 @@
 #include <SoapySDR/Logger.hpp>
 #include <stdexcept>
 
-//lazy fix for the const call issue -- FIXME
-#define _mutex const_cast<std::mutex &>(_mutex)
-#define _sock const_cast<SoapyRPCSocket &>(_sock)
-
 /*******************************************************************
  * Constructor
  ******************************************************************/
 
 SoapyRemoteDevice::SoapyRemoteDevice(const std::string &url, const SoapySDR::Kwargs &args):
-    _logAcceptor(nullptr)
+    _logAcceptor(nullptr),
+    _defaultStreamProt("udp")
 {
     //try to connect to the remote server
-    int ret = _sock.connect(url);
+    int ret = _sock.connect(url, SOAPY_REMOTE_SOCKET_TIMEOUT_US);
     if (ret != 0)
     {
         throw std::runtime_error("SoapyRemoteDevice("+url+") -- connect FAIL: " + _sock.lastErrorMsg());
@@ -37,6 +34,10 @@ SoapyRemoteDevice::SoapyRemoteDevice(const std::string &url, const SoapySDR::Kwa
     packer & args;
     packer();
     SoapyRPCUnpacker unpacker(_sock);
+
+    //default stream protocol specified in device args
+    const auto protIt = args.find("prot");
+    if (protIt != args.end()) _defaultStreamProt = protIt->second;
 }
 
 SoapyRemoteDevice::~SoapyRemoteDevice(void)
@@ -362,6 +363,49 @@ std::complex<double> SoapyRemoteDevice::getIQBalance(const int direction, const
     return result;
 }
 
+bool SoapyRemoteDevice::hasFrequencyCorrection(const int direction, const size_t channel) const
+{
+    std::lock_guard<std::mutex> lock(_mutex);
+    SoapyRPCPacker packer(_sock);
+    packer & SOAPY_REMOTE_HAS_FREQUENCY_CORRECTION;
+    packer & char(direction);
+    packer & int(channel);
+    packer();
+
+    SoapyRPCUnpacker unpacker(_sock);
+    bool result;
+    unpacker & result;
+    return result;
+}
+
+void SoapyRemoteDevice::setFrequencyCorrection(const int direction, const size_t channel, const double value)
+{
+    std::lock_guard<std::mutex> lock(_mutex);
+    SoapyRPCPacker packer(_sock);
+    packer & SOAPY_REMOTE_SET_FREQUENCY_CORRECTION;
+    packer & char(direction);
+    packer & int(channel);
+    packer & value;
+    packer();
+
+    SoapyRPCUnpacker unpacker(_sock);
+}
+
+double SoapyRemoteDevice::getFrequencyCorrection(const int direction, const size_t channel) const
+{
+    std::lock_guard<std::mutex> lock(_mutex);
+    SoapyRPCPacker packer(_sock);
+    packer & SOAPY_REMOTE_GET_FREQUENCY_CORRECTION;
+    packer & char(direction);
+    packer & int(channel);
+    packer();
+
+    SoapyRPCUnpacker unpacker(_sock);
+    double result;
+    unpacker & result;
+    return result;
+}
+
 /*******************************************************************
  * Gain API
  ******************************************************************/
@@ -685,6 +729,21 @@ std::vector<double> SoapyRemoteDevice::listSampleRates(const int direction, cons
     return result;
 }
 
+SoapySDR::RangeList SoapyRemoteDevice::getSampleRateRange(const int direction, const size_t channel) const
+{
+    std::lock_guard<std::mutex> lock(_mutex);
+    SoapyRPCPacker packer(_sock);
+    packer & SOAPY_REMOTE_GET_SAMPLE_RATE_RANGE;
+    packer & char(direction);
+    packer & int(channel);
+    packer();
+
+    SoapyRPCUnpacker unpacker(_sock);
+    SoapySDR::RangeList result;
+    unpacker & result;
+    return result;
+}
+
 /*******************************************************************
  * Bandwidth API
  ******************************************************************/
@@ -1081,6 +1140,37 @@ unsigned SoapyRemoteDevice::readRegister(const unsigned addr) const
     return unsigned(result);
 }
 
+void SoapyRemoteDevice::writeRegisters(const std::string &name, const unsigned addr, const std::vector<unsigned> &value)
+{
+    std::lock_guard<std::mutex> lock(_mutex);
+    SoapyRPCPacker packer(_sock);
+    std::vector<size_t> val (value.begin(), value.end());
+    packer & SOAPY_REMOTE_WRITE_REGISTERS;
+    packer & name;
+    packer & int(addr);
+    packer & val;
+    packer();
+
+    SoapyRPCUnpacker unpacker(_sock);
+}
+
+std::vector<unsigned> SoapyRemoteDevice::readRegisters(const std::string &name, const unsigned addr, const size_t length) const
+{
+    std::lock_guard<std::mutex> lock(_mutex);
+    SoapyRPCPacker packer(_sock);
+    packer & SOAPY_REMOTE_READ_REGISTERS;
+    packer & name;
+    packer & int(addr);
+    packer & int(length);
+    packer();
+
+    SoapyRPCUnpacker unpacker(_sock);
+    std::vector<size_t> result;
+    unpacker & result;
+    std::vector<unsigned> res (result.begin(), result.end());
+    return res;
+}
+
 /*******************************************************************
  * Settings API
  ******************************************************************/
