@@ -1,4 +1,4 @@
-// Copyright (c) 2015-2016 Josh Blum
+// Copyright (c) 2015-2017 Josh Blum
 // SPDX-License-Identifier: BSL-1.0
 
 #include <SoapySDR/Logger.hpp>
@@ -12,10 +12,6 @@
 #include "SoapyStreamEndpoint.hpp"
 #include <algorithm> //std::min, std::find
 
-//lazy fix for the const call issue -- FIXME
-#define _mutex const_cast<std::mutex &>(_mutex)
-#define _sock const_cast<SoapyRPCSocket &>(_sock)
-
 std::vector<std::string> SoapyRemoteDevice::__getRemoteOnlyStreamFormats(const int direction, const size_t channel) const
 {
     std::lock_guard<std::mutex> lock(_mutex);
@@ -130,19 +126,23 @@ SoapySDR::Stream *SoapyRemoteDevice::setupStream(
     const int direction,
     const std::string &localFormat,
     const std::vector<size_t> &channels_,
-    const SoapySDR::Kwargs &args)
+    const SoapySDR::Kwargs &args_)
 {
     //default to channel 0 when not specified
     //the channels vector cannot be empty
     //its used for stream endpoint allocation
     auto channels = channels_;
     if (channels.empty()) channels.push_back(0);
+    SoapySDR::Kwargs args(args_); //read/write copy
 
     //use the remote device's native stream format and scale factor when the conversion is supported
     double nativeScaleFactor = 0.0;
     auto nativeFormat = this->getNativeStreamFormat(direction, channels.front(), nativeScaleFactor);
     const bool useNative = (localFormat == nativeFormat) or
         (localFormat == SOAPY_SDR_CF32 and nativeFormat == SOAPY_SDR_CS16) or
+        (localFormat == SOAPY_SDR_CF32 and nativeFormat == SOAPY_SDR_CS12) or
+        (localFormat == SOAPY_SDR_CS16 and nativeFormat == SOAPY_SDR_CS12) or
+        (localFormat == SOAPY_SDR_CS16 and nativeFormat == SOAPY_SDR_CS8) or
         (localFormat == SOAPY_SDR_CF32 and nativeFormat == SOAPY_SDR_CS8) or
         (localFormat == SOAPY_SDR_CF32 and nativeFormat == SOAPY_SDR_CU8);
 
@@ -158,13 +158,27 @@ SoapySDR::Stream *SoapyRemoteDevice::setupStream(
     const auto scaleFactorIt = args.find(SOAPY_REMOTE_KWARG_SCALAR);
     if (scaleFactorIt != args.end()) scaleFactor = std::stod(scaleFactorIt->second);
 
-    size_t mtu = SOAPY_REMOTE_DEFAULT_ENDPOINT_MTU;
+    //determine reliable stream mode with tcp or datagram mode
+    std::string prot = _defaultStreamProt;
+    const auto protIt = args.find(SOAPY_REMOTE_KWARG_PROT);
+    if (protIt != args.end()) prot = protIt->second;
+    const bool datagramMode = (prot == "udp");
+    if (prot == "udp") {}
+    else if (prot == "tcp") {}
+    else throw std::runtime_error(
+        "SoapyRemote::setupStream() protcol not supported;"
+        "expected 'udp' or 'tcp', but got '"+prot+"'");
+    args[SOAPY_REMOTE_KWARG_PROT] = prot;
+
+    size_t mtu = datagramMode?SOAPY_REMOTE_DEFAULT_ENDPOINT_MTU:SOAPY_REMOTE_SOCKET_BUFFMAX;
     const auto mtuIt = args.find(SOAPY_REMOTE_KWARG_MTU);
     if (mtuIt != args.end()) mtu = size_t(std::stod(mtuIt->second));
+    args[SOAPY_REMOTE_KWARG_MTU] = std::to_string(mtu);
 
     size_t window = SOAPY_REMOTE_DEFAULT_ENDPOINT_WINDOW;
     const auto windowIt = args.find(SOAPY_REMOTE_KWARG_WINDOW);
     if (windowIt != args.end()) window = size_t(std::stod(windowIt->second));
+    args[SOAPY_REMOTE_KWARG_WINDOW] = std::to_string(window);
 
     SoapySDR::logf(SOAPY_SDR_INFO, "SoapyRemote::setup%sStream(remoteFormat=%s, localFormat=%s, scaleFactor=%g, mtu=%d, window=%d)",
         (direction == SOAPY_SDR_RX)?"Rx":"Tx", remoteFormat.c_str(), localFormat.c_str(), scaleFactor, int(mtu), int(window));
@@ -173,6 +187,9 @@ SoapySDR::Stream *SoapyRemoteDevice::setupStream(
     ConvertTypes convertType = CONVERT_MEMCPY;
     if (localFormat == remoteFormat) convertType = CONVERT_MEMCPY;
     else if (localFormat == SOAPY_SDR_CF32 and remoteFormat == SOAPY_SDR_CS16) convertType = CONVERT_CF32_CS16;
+    else if (localFormat == SOAPY_SDR_CF32 and remoteFormat == SOAPY_SDR_CS12) convertType = CONVERT_CF32_CS12;
+    else if (localFormat == SOAPY_SDR_CS16 and remoteFormat == SOAPY_SDR_CS12) convertType = CONVERT_CS16_CS12;
+    else if (localFormat == SOAPY_SDR_CS16 and remoteFormat == SOAPY_SDR_CS8) convertType = CONVERT_CS16_CS8;
     else if (localFormat == SOAPY_SDR_CF32 and remoteFormat == SOAPY_SDR_CS8) convertType = CONVERT_CF32_CS8;
     else if (localFormat == SOAPY_SDR_CF32 and remoteFormat == SOAPY_SDR_CU8) convertType = CONVERT_CF32_CU8;
     else throw std::runtime_error(
@@ -192,28 +209,33 @@ SoapySDR::Stream *SoapyRemoteDevice::setupStream(
     const auto localNode = SoapyURL(_sock.getsockname()).getNode();
     const auto remoteNode = SoapyURL(_sock.getpeername()).getNode();
 
-    //bind the stream socket to an automatic port
-    const auto bindURL = SoapyURL("udp", localNode, "0").toString();
-    int ret = data->streamSock.bind(bindURL);
-    if (ret != 0)
+    //bind the receiver side of the sockets in datagram mode
+    std::string clientBindPort, statusBindPort;
+    if (datagramMode)
     {
-        const std::string errorMsg = data->streamSock.lastErrorMsg();
-        delete data;
-        throw std::runtime_error("SoapyRemote::setupStream("+bindURL+") -- bind FAIL: " + errorMsg);
-    }
-    SoapySDR::logf(SOAPY_SDR_INFO, "Client side stream bound to %s", data->streamSock.getsockname().c_str());
-    const auto clientBindPort = SoapyURL(data->streamSock.getsockname()).getService();
+        //bind the stream socket to an automatic port
+        const auto bindURL = SoapyURL("udp", localNode, "0").toString();
+        int ret = data->streamSock.bind(bindURL);
+        if (ret != 0)
+        {
+            const std::string errorMsg = data->streamSock.lastErrorMsg();
+            delete data;
+            throw std::runtime_error("SoapyRemote::setupStream("+bindURL+") -- bind FAIL: " + errorMsg);
+        }
+        SoapySDR::logf(SOAPY_SDR_INFO, "Client side stream bound to %s", data->streamSock.getsockname().c_str());
+        clientBindPort = SoapyURL(data->streamSock.getsockname()).getService();
 
-    //bind the status socket to an automatic port
-    ret = data->statusSock.bind(bindURL);
-    if (ret != 0)
-    {
-        const std::string errorMsg = data->statusSock.lastErrorMsg();
-        delete data;
-        throw std::runtime_error("SoapyRemote::setupStream("+bindURL+") -- bind FAIL: " + errorMsg);
+        //bind the status socket to an automatic port
+        ret = data->statusSock.bind(bindURL);
+        if (ret != 0)
+        {
+            const std::string errorMsg = data->statusSock.lastErrorMsg();
+            delete data;
+            throw std::runtime_error("SoapyRemote::setupStream("+bindURL+") -- bind FAIL: " + errorMsg);
+        }
+        SoapySDR::logf(SOAPY_SDR_INFO, "Client side status bound to %s", data->statusSock.getsockname().c_str());
+        statusBindPort = SoapyURL(data->statusSock.getsockname()).getService();
     }
-    SoapySDR::logf(SOAPY_SDR_INFO, "Client side status bound to %s", data->statusSock.getsockname().c_str());
-    const auto statusBindPort = SoapyURL(data->statusSock.getsockname()).getService();
 
     //setup the remote end of the stream
     std::lock_guard<std::mutex> lock(_mutex);
@@ -227,25 +249,53 @@ SoapySDR::Stream *SoapyRemoteDevice::setupStream(
     packer & statusBindPort;
     packer();
 
-    SoapyRPCUnpacker unpacker(_sock);
+    //for tcp mode: get the binding port here and connect to it
     std::string serverBindPort;
+    if (not datagramMode)
+    {
+        SoapyRPCUnpacker unpackerTcp(_sock);
+        unpackerTcp & serverBindPort;
+        const auto connectURL = SoapyURL(prot, remoteNode, serverBindPort).toString();
+        int ret = data->streamSock.connect(connectURL);
+        if (ret != 0)
+        {
+            const std::string errorMsg = data->streamSock.lastErrorMsg();
+            delete data;
+            throw std::runtime_error("SoapyRemote::setupStream("+connectURL+") -- connect FAIL: " + errorMsg);
+        }
+        ret = data->statusSock.connect(connectURL);
+        if (ret != 0)
+        {
+            const std::string errorMsg = data->statusSock.lastErrorMsg();
+            delete data;
+            throw std::runtime_error("SoapyRemote::setupStream("+connectURL+") -- connect FAIL: " + errorMsg);
+        }
+    }
+
+    //and wait for the response with binding port and stream id
+    SoapyRPCUnpacker unpacker(_sock);
     unpacker & data->streamId;
     unpacker & serverBindPort;
 
-    //connect the stream socket to the specified port
-    const auto connectURL = SoapyURL("udp", remoteNode, serverBindPort).toString();
-    ret = data->streamSock.connect(connectURL);
-    if (ret != 0)
+    //connect the sending end of the stream socket
+    if (datagramMode)
     {
-        const std::string errorMsg = data->streamSock.lastErrorMsg();
-        delete data;
-        throw std::runtime_error("SoapyRemote::setupStream("+connectURL+") -- connect FAIL: " + errorMsg);
+        //connect the stream socket to the specified port
+        const auto connectURL = SoapyURL(prot, remoteNode, serverBindPort).toString();
+        int ret = data->streamSock.connect(connectURL);
+        if (ret != 0)
+        {
+            const std::string errorMsg = data->streamSock.lastErrorMsg();
+            delete data;
+            throw std::runtime_error("SoapyRemote::setupStream("+connectURL+") -- connect FAIL: " + errorMsg);
+        }
+        SoapySDR::logf(SOAPY_SDR_INFO, "Client side stream connected to %s", data->streamSock.getpeername().c_str());
     }
-    SoapySDR::logf(SOAPY_SDR_INFO, "Client side stream connected to %s", data->streamSock.getpeername().c_str());
 
     //create endpoint
     data->endpoint = new SoapyStreamEndpoint(data->streamSock, data->statusSock,
-        direction == SOAPY_SDR_RX, channels.size(), SoapySDR::formatToSize(remoteFormat), mtu, window);
+        datagramMode, direction == SOAPY_SDR_RX, channels.size(),
+        SoapySDR::formatToSize(remoteFormat), mtu, window);
 
     return (SoapySDR::Stream *)data;
 }
