@@ -1,4 +1,4 @@
-// Copyright (c) 2015-2016 Josh Blum
+// Copyright (c) 2015-2017 Josh Blum
 // Copyright (c) 2016-2016 Bastille Networks
 // SPDX-License-Identifier: BSL-1.0
 
@@ -63,8 +63,8 @@ bool SoapyClientHandler::handleOnce(void)
     if (not _sock.selectRecv(SOAPY_REMOTE_SOCKET_TIMEOUT_US)) return true;
 
     //receive the client's request
-    SoapyRPCUnpacker unpacker(_sock);
-    SoapyRPCPacker packer(_sock);
+    SoapyRPCUnpacker unpacker(_sock, true, -1/*no timeout*/);
+    SoapyRPCPacker packer(_sock, unpacker.remoteRPCVersion());
 
     //handle the client's request
     bool again = true;
@@ -311,6 +311,11 @@ bool SoapyClientHandler::handleOnce(SoapyRPCUnpacker &unpacker, SoapyRPCPacker &
         const auto priorityIt = args.find(SOAPY_REMOTE_KWARG_PRIORITY);
         if (priorityIt != args.end()) priority = std::stod(priorityIt->second);
 
+        std::string prot = "udp";
+        const auto protIt = args.find(SOAPY_REMOTE_KWARG_PROT);
+        if (protIt != args.end()) prot = protIt->second;
+        const bool datagramMode = (prot == "udp");
+
         //create stream
         auto stream = _dev->setupStream(direction, format, channels, args);
 
@@ -327,43 +332,82 @@ bool SoapyClientHandler::handleOnce(SoapyRPCUnpacker &unpacker, SoapyRPCPacker &
         const auto localNode = SoapyURL(_sock.getsockname()).getNode();
         const auto remoteNode = SoapyURL(_sock.getpeername()).getNode();
 
-        //bind the stream socket to an automatic port
-        const auto bindURL = SoapyURL("udp", localNode, "0").toString();
-        int ret = data.streamSock.bind(bindURL);
-        if (ret != 0)
-        {
-            const std::string errorMsg = data.streamSock.lastErrorMsg();
-            _streamData.erase(data.streamId);
-            throw std::runtime_error("SoapyRemote::setupStream("+bindURL+") -- bind FAIL: " + errorMsg);
-        }
-        SoapySDR::logf(SOAPY_SDR_INFO, "Server side stream bound to %s", data.streamSock.getsockname().c_str());
-        const auto serverBindPort = SoapyURL(data.streamSock.getsockname()).getService();
+        const auto bindURL = SoapyURL(prot, localNode, "0").toString();
+        std::string serverBindPort;
 
-        //connect the stream socket to the specified port
-        auto connectURL = SoapyURL("udp", remoteNode, clientBindPort).toString();
-        ret = data.streamSock.connect(connectURL);
-        if (ret != 0)
+        //in udp mode connect to the bound sockets on the client side
+        if (datagramMode)
         {
-            const std::string errorMsg = data.streamSock.lastErrorMsg();
-            _streamData.erase(data.streamId);
-            throw std::runtime_error("SoapyRemote::setupStream("+connectURL+") -- connect FAIL: " + errorMsg);
-        }
-        SoapySDR::logf(SOAPY_SDR_INFO, "Server side stream connected to %s", data.streamSock.getpeername().c_str());
+            data.streamSock = new SoapyRPCSocket();
+            data.statusSock = new SoapyRPCSocket();
 
-        //connect the status socket to the specified port
-        connectURL = SoapyURL("udp", remoteNode, statusBindPort).toString();
-        ret = data.statusSock.connect(connectURL);
-        if (ret != 0)
+            //bind the stream socket to an automatic port
+            int ret = data.streamSock->bind(bindURL);
+            if (ret != 0)
+            {
+                const std::string errorMsg = data.streamSock->lastErrorMsg();
+                _streamData.erase(data.streamId);
+                throw std::runtime_error("SoapyRemote::setupStream("+bindURL+") -- bind FAIL: " + errorMsg);
+            }
+            SoapySDR::logf(SOAPY_SDR_INFO, "Server side stream bound to %s", data.streamSock->getsockname().c_str());
+            serverBindPort = SoapyURL(data.streamSock->getsockname()).getService();
+
+            //connect the stream socket to the specified port
+            auto connectURL = SoapyURL("udp", remoteNode, clientBindPort).toString();
+            ret = data.streamSock->connect(connectURL);
+            if (ret != 0)
+            {
+                const std::string errorMsg = data.streamSock->lastErrorMsg();
+                _streamData.erase(data.streamId);
+                throw std::runtime_error("SoapyRemote::setupStream("+connectURL+") -- connect FAIL: " + errorMsg);
+            }
+            SoapySDR::logf(SOAPY_SDR_INFO, "Server side stream connected to %s", data.streamSock->getpeername().c_str());
+
+            //connect the status socket to the specified port
+            connectURL = SoapyURL("udp", remoteNode, statusBindPort).toString();
+            ret = data.statusSock->connect(connectURL);
+            if (ret != 0)
+            {
+                const std::string errorMsg = data.statusSock->lastErrorMsg();
+                _streamData.erase(data.streamId);
+                throw std::runtime_error("SoapyRemote::setupStream("+connectURL+") -- connect FAIL: " + errorMsg);
+            }
+            SoapySDR::logf(SOAPY_SDR_INFO, "Server side status connected to %s", data.statusSock->getpeername().c_str());
+        }
+
+        //in tcp mode, setup the server socket to listen,
+        //send the binding port back to the client and
+        //accept the client's new connections
+        else
         {
-            const std::string errorMsg = data.statusSock.lastErrorMsg();
-            _streamData.erase(data.streamId);
-            throw std::runtime_error("SoapyRemote::setupStream("+connectURL+") -- connect FAIL: " + errorMsg);
+            SoapyRPCSocket serverSocket;
+            int ret = serverSocket.bind(bindURL);
+            if (ret != 0)
+            {
+                const std::string errorMsg = serverSocket.lastErrorMsg();
+                _streamData.erase(data.streamId);
+                throw std::runtime_error("SoapyRemote::setupStream("+bindURL+") -- bind FAIL: " + errorMsg);
+            }
+            SoapySDR::logf(SOAPY_SDR_INFO, "Server side stream bound to %s", serverSocket.getsockname().c_str());
+            serverBindPort = SoapyURL(serverSocket.getsockname()).getService();
+
+            serverSocket.listen(2);
+            SoapyRPCPacker packerTcp(_sock);
+            packerTcp & serverBindPort;
+            packerTcp();
+            data.streamSock = serverSocket.accept();
+            data.statusSock = serverSocket.accept();
+            if (data.streamSock == nullptr or data.statusSock == nullptr)
+            {
+                const std::string errorMsg = serverSocket.lastErrorMsg();
+                throw std::runtime_error("SoapyRemote::setupStream("+bindURL+") -- accept FAIL: " + errorMsg);
+            }
         }
-        SoapySDR::logf(SOAPY_SDR_INFO, "Server side status connected to %s", data.statusSock.getpeername().c_str());
 
         //create endpoint
-        data.endpoint = new SoapyStreamEndpoint(data.streamSock, data.statusSock,
-            direction == SOAPY_SDR_TX, channels.size(), SoapySDR::formatToSize(format), mtu, window);
+        data.endpoint = new SoapyStreamEndpoint(*data.streamSock, *data.statusSock,
+            datagramMode, direction == SOAPY_SDR_TX, channels.size(),
+            SoapySDR::formatToSize(format), mtu, window);
 
         //start worker thread, this is not backwards,
         //receive from device means using a send endpoint
@@ -568,6 +612,54 @@ bool SoapyClientHandler::handleOnce(SoapyRPCUnpacker &unpacker, SoapyRPCPacker &
         packer & _dev->getIQBalance(direction, channel);
     } break;
 
+    ////////////////////////////////////////////////////////////////////
+    case SOAPY_REMOTE_HAS_FREQUENCY_CORRECTION:
+    ////////////////////////////////////////////////////////////////////
+    {
+        char direction = 0;
+        int channel = 0;
+        unpacker & direction;
+        unpacker & channel;
+        #ifdef SOAPY_SDR_API_HAS_FREQUENCY_CORRECTION_API
+        packer & _dev->hasFrequencyCorrection(direction, channel);
+        #else
+        bool result(false);
+        packer & result;
+        #endif
+    } break;
+
+    ////////////////////////////////////////////////////////////////////
+    case SOAPY_REMOTE_SET_FREQUENCY_CORRECTION:
+    ////////////////////////////////////////////////////////////////////
+    {
+        char direction = 0;
+        int channel = 0;
+        double value(0.0);
+        unpacker & direction;
+        unpacker & channel;
+        unpacker & value;
+        #ifdef SOAPY_SDR_API_HAS_FREQUENCY_CORRECTION_API
+        _dev->setFrequencyCorrection(direction, channel, value);
+        #endif
+        packer & SOAPY_REMOTE_VOID;
+    } break;
+
+    ////////////////////////////////////////////////////////////////////
+    case SOAPY_REMOTE_GET_FREQUENCY_CORRECTION:
+    ////////////////////////////////////////////////////////////////////
+    {
+        char direction = 0;
+        int channel = 0;
+        unpacker & direction;
+        unpacker & channel;
+        #ifdef SOAPY_SDR_API_HAS_FREQUENCY_CORRECTION_API
+        packer & _dev->getFrequencyCorrection(direction, channel);
+        #else
+        double result(0.0);
+        packer & result;
+        #endif
+    } break;
+
     ////////////////////////////////////////////////////////////////////
     case SOAPY_REMOTE_LIST_GAINS:
     ////////////////////////////////////////////////////////////////////
@@ -834,6 +926,22 @@ bool SoapyClientHandler::handleOnce(SoapyRPCUnpacker &unpacker, SoapyRPCPacker &
         packer & _dev->listSampleRates(direction, channel);
     } break;
 
+    ////////////////////////////////////////////////////////////////////
+    case SOAPY_REMOTE_GET_SAMPLE_RATE_RANGE:
+    ////////////////////////////////////////////////////////////////////
+    {
+        char direction = 0;
+        int channel = 0;
+        unpacker & direction;
+        unpacker & channel;
+        #ifdef SOAPY_SDR_API_HAS_GET_SAMPLE_RATE_RANGE
+        packer & _dev->getSampleRateRange(direction, channel);
+        #else
+        SoapySDR::RangeList result;
+        packer & result;
+        #endif
+    } break;
+
     ////////////////////////////////////////////////////////////////////
     case SOAPY_REMOTE_SET_BANDWIDTH:
     ////////////////////////////////////////////////////////////////////
@@ -1127,6 +1235,43 @@ bool SoapyClientHandler::handleOnce(SoapyRPCUnpacker &unpacker, SoapyRPCPacker &
         #endif
     } break;
 
+    ////////////////////////////////////////////////////////////////////
+    case SOAPY_REMOTE_WRITE_REGISTERS:
+    ////////////////////////////////////////////////////////////////////
+    {
+        std::string name;
+        int addr = 0;
+        std::vector<size_t> value;
+        unpacker & name;
+        unpacker & addr;
+        unpacker & value;
+        #ifdef SOAPY_SDR_API_HAS_NAMED_REGISTERS_API
+        std::vector <unsigned> val (value.begin(), value.end());
+        _dev->writeRegisters(name, unsigned(addr), val);
+        #endif
+        packer & SOAPY_REMOTE_VOID;
+    } break;
+
+    ////////////////////////////////////////////////////////////////////
+    case SOAPY_REMOTE_READ_REGISTERS:
+    ////////////////////////////////////////////////////////////////////
+    {
+        std::string name;
+        int addr = 0;
+        int length;
+        unpacker & name;
+        unpacker & addr;
+        unpacker & length;
+        #ifdef SOAPY_SDR_API_HAS_NAMED_REGISTERS_API
+        std::vector <unsigned> val = _dev->readRegisters(name, unsigned(addr), size_t(length));
+        std::vector <size_t> value (val.begin(), val.end());
+        packer & (value);
+        #else
+        std::vector <size_t> value;
+        packer & (value);
+        #endif
+    } break;
+
     ////////////////////////////////////////////////////////////////////
     case SOAPY_REMOTE_GET_SETTING_INFO:
     ////////////////////////////////////////////////////////////////////
